import { Schema } from "../schema/classes/Schema";
import Api, { RQLParams, ResultSet, CurrentUser } from "./Api";
import Transaction from "./Transaction";

/**
 * Class used to communicate with a CubicWeb instance's API.
 */
export default class Client {
  private api: Api;

  /**
   * Class used to communicate with a CubicWeb instance's API.
   *
   * @param apiUrl The API url. It should not include the API version as it will be automatically added.
   * Example: if you can retrieve the schema with https://example.com/api/v1/schema,
   * use https://example.com/api as `apiUrl`.
   */
  constructor(apiUrl: string) {
    this.api = new Api(apiUrl);
  }

  get apiUrl() {
    return this.api.apiUrl;
  }

  /**
   * Executes the given query and waits for the response.
   *
   * @param query The RQL query to send
   * @param params The additional parameters for the request
   * @returns A promise with the request's result set
   */
  async execute(query: string, params: RQLParams = {}): Promise<ResultSet> {
    return await this.api.execute(query, params);
  }

  /**
   * Retrieves the CubicWeb instance's JSON schema.
   *
   * @returns The CubicWeb instance's JSON schema.
   */
  async schema() {
    const schema = await this.api.schema();
    return new Schema(schema);
  }

  /**
   * Tries to log in the user on the configured CubicWeb instance with the given credentials.
   *
   * @param login The user login
   * @param password THe user password
   * @returns A promise returning true if the login was successful, false otherwise
   */
  async login(login: string, password: string) {
    return await this.api.login(login, password);
  }

  /**
   * Get the user associated to this client
   *
   * @returns A promise resolving to the current user
   */
  async currentUser(): Promise<CurrentUser> {
    return await this.api.currentUser();
  }

  /**
   * Creates a new transaction and exposes the Transaction object in the given callback.
   * Use this callback to execute RQL requests as part of the transaction.
   * Once the callback's end is reached, the transaction is automatically committed.
   * Note that any exception thrown by the callback will lead to the
   * transaction being rolled back.
   *
   * @param callback Function to execute transaction requests in.
   * @returns The promise returned by the callback
   */
  async transaction<T>(callback: (tr: Transaction) => Promise<T>): Promise<T> {
    const transaction = new Transaction(this.api);
    await transaction.begin();

    let result: T;
    try {
      result = await callback(transaction);
    } catch (error) {
      if (!transaction.isClosed()) {
        await transaction.rollback();
      }
      throw error;
    }
    if (!transaction.isClosed()) {
      await transaction.commit();
    }
    return result;
  }
}
