import { TransactionClosedError } from "../Errors";
import Api, { RQLParams } from "./Api";

/**
 * Class used to handle CubicWeb's transactions.
 * A transaction allows to execute several RQL requests without altering the database until all
 * changes are committed.
 *
 * You should never have to create this object yourself.
 * Instead, retrieve this object from the methods in the `Client` class.
 */
export default class Transaction {
  private _uuid: string | null = null;
  private _closed = false;

  /**
   * Class used to handle CubicWeb's transactions.
   * A transaction allows to execute several RQL requests without altering the database until all
   * changes are committed.
   *
   * You should never have to create this object yourself.
   * Instead, retrieve this object from the methods in the `Client` class.
   *
   * @param api The API object used to communicate with the CubicWeb backend.
   */
  constructor(private api: Api) {}

  /**
   * @returns True if the connection is closed.
   */
  isClosed() {
    return this._closed;
  }

  /**
   * This transaction unique identifier.
   */
  get uuid() {
    return this._uuid;
  }

  /**
   * Begins a new transaction an retrieves the associated uuid.
   */
  async begin() {
    this._closed = false;
    this._uuid = await this.api.transactionBegin();
  }

  /**
   * Executes the given query and waits for the response.
   * Make sure you called `begin` before using this method.
   * As this is a transaction changes will only be applied when you call the `commit` method.
   *
   * @param query The RQL query to send
   * @param params The additional parameters for the request
   * @returns A promise with the request's result set
   */
  async execute(query: string, params: RQLParams = {}) {
    this.preventUsageIfClosed();
    return await this.api.transactionExecute(this.uuid, query, params);
  }

  /**
   * Commits the transaction to apply all the RQL queries sent with `execute`.
   */
  async commit() {
    this.preventUsageIfClosed();
    await this.api.transactionCommit(this.uuid);
    this._closed = true;
  }

  /**
   * Rollbacks all the queries sent with `execute` and closes the transaction.
   */
  async rollback() {
    this.preventUsageIfClosed();
    await this.api.transactionRollback(this.uuid);
    this._closed = true;
  }

  private preventUsageIfClosed(): asserts this is { uuid: string } {
    if (this.isClosed()) {
      throw new TransactionClosedError();
    }
  }
}
