import { EntityRawSchemaArray } from "../raw/Entities";
import {
  Cardinality,
  RelationDefinitionRawSchema,
  RelationDefinitionRawSchemaArray,
} from "../raw/RelationDefinitions";
import { Schema } from "./Schema";
import { EntitySchema } from "./EntitySchema";

/**
 * Returns the given schema using loose typing.
 * Can be useful when working with generic schemas
 *
 * @param schema The schema to create a loose type for
 * @returns A loosely typed schema
 */
export const getLooseSchema = <
  E extends EntityRawSchemaArray = EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E> = RelationDefinitionRawSchemaArray<E>
>(
  schema: Schema<E, R>
) => {
  return schema as unknown as Schema;
};

/**
 * Checks if the given cardinality requires at least one entity to be present
 *
 * @param cardinality The cardinality to check
 * @returns True if the cardinality is 1 or +
 */
export function cardinalityRequired(cardinality: Cardinality): boolean {
  return cardinality === "1" || cardinality === "+";
}

/**
 * Partial raw relation definition schema type.
 */
export type RelationDefinitionMatcher<E extends EntityRawSchemaArray> = Partial<
  RelationDefinitionRawSchema<E>
>;

/**
 * Type representing a raw relation definition matched from a partial raw relation definition schema.
 * If it finds a relation definition matching the partial one given, it will return its type.
 * If it cannot find any relation definition matching, it will return the `never` type
 * and TypeScript will give compile errors.
 */
export type MatchedRelationDefinition<
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<EntityRawSchemaArray>,
  M extends RelationDefinitionMatcher<E>
> = R[number] extends infer RDef ? (RDef extends M ? RDef : never) : never;

export function hasAttributeWithName(
  entitySchema: EntitySchema,
  attributeNameSearched: string
): boolean {
  for (const entityAttribute of entitySchema.attributes) {
    if (entityAttribute.type == attributeNameSearched) {
      return true;
    }
  }
  return false;
}

export function isAttributeRichString(
  entitySchema: EntitySchema,
  attributeNameSearched: string
): boolean {
  return hasAttributeWithName(entitySchema, attributeNameSearched + "_format");
}
