/**
 * Base type used to describe a constraint.
 */
type ConstraintBase<
  T extends string,
  V extends Record<string, unknown> | null
> = {
  type: T;
  value: V;
};

/**
 * All the possible operators inside a constraint.
 */
export type ConstraintOperator = ">" | ">=" | "<" | "<=";

/**
 * THe type of a constraint message when one is present.
 */
type ConstraintMessage = string | null;

export type UniqueConstraint = ConstraintBase<
  "UniqueConstraint",
  { msg: ConstraintMessage }
>;

export type SizeConstraint = ConstraintBase<
  "SizeConstraint",
  { min: number | null; max: number | null; msg: ConstraintMessage }
>;

export type RegexpConstraint = ConstraintBase<
  "RegexpConstraint",
  { regexp: string; flags: number; msg: ConstraintMessage }
>;

export type BoundaryConstraint = ConstraintBase<
  "BoundaryConstraint",
  { boundary: number | null; op: ConstraintOperator; msg: ConstraintMessage }
>;
export type IntervalBoundConstraint = ConstraintBase<
  "IntervalBoundConstraint",
  { minvalue: number | null; maxvalue: number | null; msg: ConstraintMessage }
>;

export type StaticVocabularyConstraint = ConstraintBase<
  "StaticVocabularyConstraint",
  { values: string[]; msg: ConstraintMessage }
>;

export type FormatConstraint = ConstraintBase<
  "FormatConstraint",
  {
    values: ["text/rest", "text/markdown", "text/html", "text/plain"];
    msg: ConstraintMessage;
  }
>;

export type RQLVocabularyConstraint = ConstraintBase<
  "RQLVocabularyConstraint",
  { expression: string; mainvars: string[] }
>;

export type RQLConstraint = ConstraintBase<
  "RQLConstraint",
  { expression: string; mainvars: string[]; msg: ConstraintMessage }
>;

/**
 * Type representing all possible constraints for attributes.
 */
export type AttributeConstraint = Readonly<
  | UniqueConstraint
  | SizeConstraint
  | RegexpConstraint
  | BoundaryConstraint
  | IntervalBoundConstraint
  | StaticVocabularyConstraint
  | FormatConstraint
  | RQLConstraint
>;

/**
 * Type representing all possible constraints for relations.
 */
export type RelationConstraint = Readonly<
  RQLVocabularyConstraint | RQLConstraint
>;
