import {
  Transaction,
  Schema,
  EntitySchema,
  EntityRawSchemaArray,
  RelationDefinitionRawSchemaArray,
  getLooseSchema,
  RelationDefinitionSchema,
} from "@cubicweb/client";
import { Identifier, RaRecord } from "../types";
import HttpError from "../HttpError";
import {
  getRelationVariable,
  initVariableKeyAssociation,
  VariableKeyAssociation,
} from "./constants";
import { getRestrictionsClause } from "./getParams";
import { apiErrorToReactAdmin } from "./requests";
import { getEntitySchemaNames, getAttributesNames } from "./schema";
import {
  CubicWebReferenceArrayInputValueType,
  getRelationFromKey,
  isIdentifier,
} from "./relations";

/**
 * Type representing a new entity to create.
 * It contains the type of the entity, the relation key the created entity should be added to
 * and the data the entity should be created with.
 */
type NewEntitiesToCreate = {
  type: string;
  relation: string;
  data: FormData;
};

/**
 * Type representing an object with the relations as keys, and the ids
 * of new entities created as values.
 */
type NewEntitiesCreated = Record<string, number[]>;

/**
 * Type representing the list of RQL expressions to use as assignements and restrictions
 * for the final RQL query.
 */
type AssignementsAndRestrictionsRqlParts = {
  assignments: string[];
  restrictions: string[];
};

type FormData = Partial<RaRecord>;

/**
 * Creates the assignements and the restrictions parts of the RQL query to build new relations.
 *
 * The generated RQL query parts will link the given ids to the entity we are creating
 * using the given relation definition, as subject or object.
 *
 * @param relation The relation definition schema to use to link the ids
 * @param idsRole If the ids are a subject or object in the relation
 * @param ids An id or the list of ids in the relation
 * @returns An object containing assignements and relations arrays
 */
export const generateRqlCreateRelationsParts = (
  relation: RelationDefinitionSchema,
  variableKeyAssociation: VariableKeyAssociation,
  idsRole: "subject" | "object",
  ids: Identifier[] | Identifier | undefined
): AssignementsAndRestrictionsRqlParts => {
  const params: AssignementsAndRestrictionsRqlParts = {
    assignments: [],
    restrictions: [],
  };
  const targetEntityVariable = getRelationVariable(
    getRelationKey(relation, idsRole),
    "relation",
    variableKeyAssociation
  );
  const idsList = ids != null ? (Array.isArray(ids) ? ids : [ids]) : [];
  idsList.forEach((id, index) => {
    if (idsRole === "subject") {
      params.assignments.push(
        `X ${relation.type} ${targetEntityVariable}${index}`
      );
    } else {
      params.assignments.push(
        `${targetEntityVariable}${index} ${relation.type} X`
      );
    }
    params.restrictions.push(`${targetEntityVariable}${index} eid ${id}`);
  });
  return params;
};

/**
 * Extracts the list of new entities to create from the given input data.
 *
 * @param inputData Input data to find elements to create in
 * @returns
 */
export const GetEntitiesToCreateFromInput = (
  inputData:
    | number
    | FormData
    | CubicWebReferenceArrayInputValueType
    | undefined
): FormData[] => {
  // The form data is an existing entity, ignore it
  if (typeof inputData === "number") {
    return [];
  }
  // The form data is from a CubicWebReferenceArrayInput
  // In this case we use the value from the created key
  if (inputData?.created != null && Array.isArray(inputData.created)) {
    return inputData.created;
  }
  // The form data is from a CubicWebReferenceInput
  // In this case we use it as it is
  if (inputData != null && inputData.id !== "temp_input") {
    return [inputData];
  }
  // The form data is something else, ignore it
  return [];
};

/**
 * Retrieves a list of child entities to create from the form data.
 *
 * These child entities must be created first in order to create the parent entity.
 *
 * @param entitySchema The schema of the entity we are creating
 * @param formData The form data used to create the entity
 * @returns An array of new entities to create
 */
export const getChildEntitiesToCreate = (
  entitySchema: EntitySchema | null,
  formData: FormData
): NewEntitiesToCreate[] => {
  const list: NewEntitiesToCreate[] = [];
  const roles = ["subject", "object"] as const;
  if (entitySchema) {
    const relationsObject = {
      subject: entitySchema.subjectRelationDefinitions,
      object: entitySchema.objectRelationsDefinitions,
    };
    roles.forEach((role) => {
      const relations = relationsObject[role];
      relations.forEach((rel) => {
        const relationKey = getRelationKey(rel, role);
        let newElements: FormData[] = [];
        const inputData = formData[relationKey];
        if (Array.isArray(inputData)) {
          newElements = newElements.concat(
            inputData.flatMap((e) => GetEntitiesToCreateFromInput(e))
          );
        } else {
          newElements = newElements.concat(
            GetEntitiesToCreateFromInput(inputData)
          );
        }
        newElements.forEach((e) =>
          list.push({
            type: role === "subject" ? rel.object.type : rel.subject.type,
            relation: relationKey,
            data: e,
          })
        );
      });
    });
  }

  return list;
};

/**
 * Recursivly creates the given new entities.
 *
 * @param schema The CubicWeb schema
 * @param transaction The transaction object to use when making the requests
 * @param newElementsToCreate The array of new elements to create
 * @returns A promise of an object containing the ids of the created entities for each relation
 */
export const createNewEntities = async <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(
  schema: Schema<E, R>,
  transaction: Transaction,
  newElementsToCreate: NewEntitiesToCreate[]
): Promise<NewEntitiesCreated> => {
  const newIds: NewEntitiesCreated = {};
  for (const newSubject of newElementsToCreate) {
    const result = await makeRecursiveCreateRequest(
      schema,
      newSubject.type,
      newSubject.data,
      transaction
    );
    if (result != null) {
      if (newIds[newSubject.relation] == null) {
        newIds[newSubject.relation] = [result.id as number];
      } else {
        newIds[newSubject.relation].push(result.id as number);
      }
    }
  }
  return newIds;
};

/**
 * Replaces the data used to create child entities by their ids.
 * To get the ids, use the `createNewEntities` function.
 *
 * The resulting object is then ready to create the parent entity.
 *
 * @param formData The original data containing entities to create
 * @param newEntities The new entities created
 * @returns AN object with entity data replaced by their ids
 */
export const mergeNewElementsIntoData = (
  formData: FormData,
  newEntities: NewEntitiesCreated
): FormData => {
  const newData: FormData = {};
  Object.entries(formData).forEach(([key, value]) => {
    if (Object.keys(newEntities).includes(key)) {
      if (value != null && value.id === "temp_input") {
        if (value.selected != null && Array.isArray(value.selected)) {
          newData[key] = {
            ...value,
            selected: [...value.selected, ...newEntities[key]],
          };
        } else {
          newData[key] = {
            ...value,
            selected: [...newEntities[key]],
          };
        }
      }
    } else {
      newData[key] = value;
    }
  });
  return newData;
};

/**
 * Recursivly finds and creates child entities from the form data.
 * The data to create the child entities is then replaced by their ids.
 *
 * @param schema The CubicWeb schema
 * @param entitySchema The schema for the current entity
 * @param transaction The transaction object to use when making the requests
 * @param formData The form data containing the data for the entity
 * @returns A promise of an object containing the form data with the child entities ids
 * instead of their data.
 */
export const findAndCreateChildEntities = async <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(
  schema: Schema<E, R>,
  entitySchema: EntitySchema,
  transaction: Transaction,
  formData: FormData
): Promise<FormData> => {
  const newElementsToCreate = getChildEntitiesToCreate(entitySchema, formData);
  const createdElements = await createNewEntities(
    schema,
    transaction,
    newElementsToCreate
  );
  return mergeNewElementsIntoData(formData, createdElements);
};

/**
 * Generates all the assignements and restrictions parts of the RQL query to create a new entity.
 *
 * @param entitySchema The schema for the current entity
 * @param formData The form data used to create the entity
 */
export const generateRqlCreateParts = (
  entitySchema: EntitySchema,
  formData: FormData,
  variableKeyAssociation: VariableKeyAssociation
): AssignementsAndRestrictionsRqlParts => {
  const params: AssignementsAndRestrictionsRqlParts = {
    assignments: [],
    restrictions: [],
  };

  const attributesNamesList = getAttributesNames(entitySchema);
  for (const [key, value] of Object.entries(formData)) {
    if (value !== undefined && attributesNamesList.includes(key)) {
      params.assignments.push(`X ${key} %(${key})s`);
    }
  }

  const availableRoles = ["subject", "object"] as const;
  availableRoles.forEach((role) => {
    getEntitySchemaNames(entitySchema, `${role}_relations`).forEach(
      (relationName) => {
        const relationData = formData[relationName] as
          | Identifier
          | CubicWebReferenceArrayInputValueType
          | undefined;
        if (relationData === undefined) {
          return;
        }
        const relation = getRelationFromKey(relationName, entitySchema);
        const result = generateRqlCreateRelationsParts(
          relation,
          variableKeyAssociation,
          role,
          isIdentifier(relationData) || relationData === null
            ? relationData
            : relationData.selected
        );
        params.assignments.push(...result.assignments);
        params.restrictions.push(...result.restrictions);
      }
    );
  });
  return params;
};

/**
 * Creates a entity using the form data, and recursively any child entity present in the form data.
 *
 * @param schema The CubicWeb schema
 * @param resource The entity type
 * @param formData The form data used to create the entity
 * @param transaction The transaction object to use when making the requests
 */
export const makeRecursiveCreateRequest = async <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(
  schema: Schema<E, R>,
  resource: string,
  formData: FormData,
  transaction: Transaction
) => {
  const entitySchema = getLooseSchema(schema).getEntity(resource);
  if (entitySchema === null) {
    throw new RangeError(`Entity '${resource}' does not exist in the schema`);
  }
  const variableKeyAssociation: VariableKeyAssociation =
    initVariableKeyAssociation();
  // Recursive call
  const newData = await findAndCreateChildEntities(
    getLooseSchema(schema),
    entitySchema,
    transaction,
    formData
  );

  const { assignments, restrictions } = generateRqlCreateParts(
    entitySchema,
    newData,
    variableKeyAssociation
  );

  const requestStr = `INSERT ${resource} X: ${assignments.join(
    ", "
  )} ${getRestrictionsClause(restrictions)}`;

  const rows = await transaction.execute(requestStr, formData);
  if (rows == null || rows[0][0] == null) {
    throw new Error(
      "Could not understand server response. Expected and array of identifiers."
    );
  }

  const id = rows[0][0] as Identifier;
  return {
    ...removeRelationData(entitySchema, newData),
    id,
    eid: id,
  } as RaRecord;
};

const RELATION_SEPARATOR = "__";
const REVERSE_INDICATOR = "reverse_";

export type Cardinality = "*" | "1" | "?" | "+";

/**
 * @param cardinality The cardinility to check
 * @returns True if the cardinality is `*` or `+`
 *
 * @category Relations Utilities
 */
export const isCardinalityMulitple = (cardinality: Cardinality) => {
  return cardinality === "*" || cardinality === "+";
};

export const removeRelationData = (
  entitySchema: EntitySchema,
  data: Partial<RaRecord>
) => {
  const relationToIgnore = [
    ...entitySchema.subjectRelationDefinitions
      .filter((r) => isCardinalityMulitple(r.objectCardinality))
      .map((r) => getRelationKey(r, "subject")),
    ...entitySchema.objectRelationsDefinitions
      .filter((r) => isCardinalityMulitple(r.subjectCardinality))
      .map((r) => getRelationKey(r, "object")),
  ];
  const newData: Partial<RaRecord> = {};
  Object.entries(data).forEach(([key, value]) => {
    if (!relationToIgnore.includes(key)) {
      newData[key] = value;
    }
  });
  return newData;
};

/**
 * Gets a unique key representing the tuple: mode - subject - relation type - object
 *
 * Reverse operation from `getRelationFromKey`.
 *
 * @param relation The relation definition to get a key for
 * @param role Subject or object mroe
 * @returns A unique key representing the relation definition
 *
 * @category Relations Utilities
 */
export const getRelationKey = (
  relation: Readonly<RelationDefinitionSchema>,
  role: "subject" | "object"
) => {
  const type = `${relation.subject.type}${RELATION_SEPARATOR}${relation.type}${RELATION_SEPARATOR}${relation.object.type}`;
  return role === "subject" ? type : `${REVERSE_INDICATOR}${type}`;
};
