import {
  CardinalityPair,
  EntitySchema,
  RelationDefinitionSchema,
} from "@cubicweb/client";
import { Identifier } from "../types";
import {
  getRelationFromKey,
  getRelationKey,
  isCardinalityMulitple,
  isRelationReversed,
} from "./relations";
import {
  AttributesRqlRequestParamsType,
  getRelationVariable,
  initVariableKeyAssociation,
  RelationsRqlRequestParamsType,
  RESTRICTED_PREFIX_LIST,
  SelectionTransformer,
  SortJsonType,
  VariableKeyAssociation,
} from "./constants";
import { getAttributesNames, getCompatibleRelations } from "./schema";

/**
 * Type representing a list of filters compatible with CubicWeb relations.
 */
export interface CubicWebFilter {
  relations?: Record<string, CubicWebRelationFilter>;
  rql?: string[];
}

/**
 * Type representing a filter compatible with CubicWeb relations.
 */
export interface CubicWebRelationFilter {
  // The ids to filter against
  ids?: {
    include?: Identifier[];
    exclude?: Identifier[];
  };
  // The string to filter against
  text?: string;
  // The name of the variable to compare against.
  // Use this to make filters relative to each other
  // Defaults to 'X'
  variableName?: string;
}

/**
 * Creates RQL query parts needed to implement a filter.
 *
 * @param filter An object containing filters to use for each attribute.
 * @returns A list of RQL query parts needed to implement a filter
 */
export const getFilterRqlParts = (
  filter: CubicWebFilter,
  entity: EntitySchema,
  variableKeyAssociation: VariableKeyAssociation
): string[] => {
  const rqlFilters = filter.rql ? filter.rql : [];
  const relationsFilters = filter.relations ? filter.relations : {};

  return [
    ...rqlFilters,
    ...Object.entries(relationsFilters)
      .map(([relationKey, relationFilter]) => {
        // Make sure the key is valid
        const relation = getRelationFromKey(relationKey, entity);
        if (relation == null) {
          return undefined;
        }

        // Let users change the main variable name from X
        const mainVarName = relationFilter.variableName
          ? relationFilter.variableName
          : "X";

        if (relation.final) {
          return getAttributeFilterRqlParts(
            relationKey,
            relationFilter,
            mainVarName
          );
        }
        return getRelationFilterRqlParts(
          relationKey,
          relationFilter,
          mainVarName,
          relation,
          variableKeyAssociation
        );
      })
      .flat(1)
      // remove invalid filters
      .filter((v): v is string => v != null),
  ];
};

/**
 * Creates RQL query parts needed to implement a filter on a relation.
 *
 * @param relationKey The relation key used as filter
 * @param relationFilter The filter to apply for this relation
 * @param mainVarName The main variable name to use in the filter.
 * @param relation The relation definition used as filter
 * @returns A list of RQL query parts
 */
const getRelationFilterRqlParts = (
  relationKey: string,
  relationFilter: CubicWebRelationFilter,
  mainVarName: string,
  relation: RelationDefinitionSchema,
  variableKeyAssociation: VariableKeyAssociation
): string[] | undefined => {
  const relationVar = `FILTER_${getRelationVariable(
    relationKey,
    "relation",
    variableKeyAssociation
  )}`;
  // Only filter relations on ids
  if (relationFilter.ids != null) {
    // Subject relations
    const filters = [];
    if (!isRelationReversed(relationKey)) {
      const typePart = `${mainVarName} ${relation.type} ${relationVar},`;
      // Include ids
      if (
        relationFilter.ids.include != null &&
        relationFilter.ids.include?.length > 0
      ) {
        filters.push(
          `EXISTS(${typePart} ${relationVar} eid IN(${relationFilter.ids.include.join(
            ", "
          )}))`
        );
      }
      // Exclude ids
      if (
        relationFilter.ids.exclude != null &&
        relationFilter.ids.exclude?.length > 0
      ) {
        filters.push(
          `NOT EXISTS(${typePart} ${relationVar} eid IN(${relationFilter.ids.exclude.join(
            ", "
          )}))`
        );
      }
      return filters;
      // Object relations
    } else {
      const typePart = `${relationVar} ${relation.type} ${mainVarName},`;
      // Include ids
      if (
        relationFilter.ids.include != null &&
        relationFilter.ids.include?.length > 0
      ) {
        filters.push(
          `EXISTS(${typePart} ${relationVar} eid IN(${relationFilter.ids.include.join(
            ", "
          )}))`
        );
      }
      // Exclude ids
      if (
        relationFilter.ids.exclude != null &&
        relationFilter.ids.exclude?.length > 0
      ) {
        filters.push(
          `NOT EXISTS(${typePart} ${relationVar} eid IN(${relationFilter.ids.exclude.join(
            ", "
          )}))`
        );
      }
      return filters;
    }
  }
};

/**
 * Creates RQL query parts needed to implement a filter on an attribute.
 *
 * @param relationKey The relation key used as filter
 * @param relationFilter The filter to apply for this attribute
 * @param mainVarName The main variable name to use in the filter.
 * @returns A list of RQL query parts
 */
const getAttributeFilterRqlParts = (
  relationKey: string,
  relationFilter: CubicWebRelationFilter,
  mainVarName: string
) => {
  const realAttrName = relationKey === "id" ? "eid" : relationKey;
  if (relationFilter.text != null && relationFilter.text !== "") {
    return `EXISTS(${mainVarName} ${realAttrName} ~= '%${relationFilter.text.replace(
      "'",
      "\\'"
    )}%')`;
    // Handle ids filters
  } else if (relationFilter.ids != null) {
    const filters = [];
    // Include ids
    if (
      relationFilter.ids.include != null &&
      relationFilter.ids.include?.length > 0
    ) {
      filters.push(
        `EXISTS(${mainVarName} ${realAttrName} IN (${relationFilter.ids.include.join(
          ", "
        )}))`
      );
    }
    // Exclude ids
    if (
      relationFilter.ids.exclude != null &&
      relationFilter.ids.exclude?.length > 0
    ) {
      filters.push(
        `NOT EXISTS(${mainVarName} ${realAttrName} IN (${relationFilter.ids.exclude.join(
          ", "
        )}))`
      );
    }
    return filters;
  } else {
    return undefined;
  }
};

/**
 * Tries to convert the given json string to a sort object.
 * It also warns the user if it finds restricted prefixes.
 *
 * @param sortString The json representation of the sort object
 * @returns The converted sort object if valid, null otherwise
 */
export const getSortJson = (sortString: string): SortJsonType | null => {
  try {
    const jsonData = JSON.parse(sortString);
    if (
      typeof jsonData["orderExpression"] === "string" &&
      (jsonData["restriction"] == null ||
        typeof jsonData["restriction"] === "string")
    ) {
      const sortData = jsonData as SortJsonType;
      const warnIndex = RESTRICTED_PREFIX_LIST.findIndex(
        (p) =>
          sortData.restriction?.includes(p) ||
          sortData.orderExpression.includes(p)
      );
      if (warnIndex != -1) {
        console.warn(
          `Found restricted prefix '${
            RESTRICTED_PREFIX_LIST[warnIndex]
          }' in sortBy RQL prop.\nPlease avoid using restricted prefixes (${RESTRICTED_PREFIX_LIST.join(
            ", "
          )}) as these are used for auto-generated variables.`
        );
      }
      return sortData;
    } else {
      return null;
    }
  } catch (e) {
    return null;
  }
};

/**
 * Tries to convert the given meta params into the appropriate
 * getAllRqlParts params.
 *
 * @param meta The meta object given to the getOne or getMany function
 * @returns A dictionnary which contains all getAllRqlParts named params
 */
export const getAllRqlPartsParamsFromMeta = (meta?: {
  resolveRelations?:
    | { [key: string]: true }
    | "all"
    | "allSubject"
    | "allObject";
}) => {
  let includeObjectRelations = false;
  let includeSubjectRelations = false;
  let resolveRelations = undefined;
  const knownRelation = undefined;
  const sortData = undefined; // TODO: add meta to sort data
  if (meta !== undefined && meta.resolveRelations !== undefined) {
    const metaFetchRelations = meta.resolveRelations;
    if (metaFetchRelations === "all") {
      includeSubjectRelations = true;
      includeObjectRelations = true;
    } else if (metaFetchRelations === "allSubject") {
      includeSubjectRelations = true;
    } else if (metaFetchRelations === "allObject") {
      includeObjectRelations = true;
    } else if (typeof metaFetchRelations === "object") {
      includeSubjectRelations = true;
      includeObjectRelations = true;
      resolveRelations = metaFetchRelations;
    }
  }
  return {
    includeObjectRelations,
    includeSubjectRelations,
    resolveRelations,
    knownRelation,
    sortData,
  };
};

/**
 * Creates all the RQL query parts needed to fetch attributes and relations of the given entity.
 *
 * @param entitySchema The entity schema abstraction class
 * @param sortData The attribute used to sort the request or a json object with a rql request
 * (see the `getRqlSortByProp` function)
 * @param includeSubjectRelations True to include relation with this entity as subject
 * @param includeObjectRelations True to include relation with this entity as object
 * @param knownRelation An object containing the type and the value for the known relation
 * @returns An object containing RQL the query parts
 */
export const getAllRqlParts = (
  entitySchema: EntitySchema,
  sortData?: string,
  includeSubjectRelations = false,
  includeObjectRelations = false,
  knownRelation?: { relation: string; value: unknown },
  resolveRelations?: { [key: string]: true }
) => {
  const variableKeyAssociation: VariableKeyAssociation =
    initVariableKeyAssociation();
  const attributesNames = getAttributesNames(entitySchema);
  const rqlAttributes = getAttributesRqlParts(
    attributesNames,
    variableKeyAssociation,
    sortData,
    knownRelation
  );
  const rqlRelations = getRelationsSubjectObjectRqlParts(
    entitySchema,
    variableKeyAssociation,
    sortData,
    includeSubjectRelations,
    includeObjectRelations,
    knownRelation,
    resolveRelations
  );

  let sortExpression =
    rqlAttributes.sortVariable != null
      ? rqlAttributes.sortVariable
      : rqlRelations.sortVariable;
  const sortRestrictions: string[] = [];
  if (sortData?.startsWith("{")) {
    const sortJson = getSortJson(sortData);
    if (sortJson != null) {
      sortExpression = sortJson.orderExpression;
      if (sortJson.restriction != null) {
        sortRestrictions.push(sortJson.restriction);
      }
    }
  }

  return {
    selectionTransformers: [
      ...rqlAttributes.selectionTransformers,
      ...rqlRelations.selectionTransformers,
    ],
    selection: [...rqlAttributes.selection, ...rqlRelations.selection],
    groupBy: [...rqlAttributes.groupBy, ...rqlRelations.groupBy],
    restrictions: [
      ...rqlAttributes.restrictions,
      ...rqlRelations.restrictions,
      ...sortRestrictions,
    ],
    relationsTypes: [...attributesNames, ...rqlRelations.relationsTypes],
    sortExpression,
  };
};

/**
 * Creates RQL query parts to fetch ids of entities linked by a relation to the given entity schema
 * as subject and object.
 *
 * @param entitySchema The entity schema abstraction class
 * @param sortRelation The relation type used to sort
 * @param includeSubjectRelations True to include relation with this entity as subject
 * @param includeObjectRelations True to include relation with this entity as object
 * @param knownRelation An object containing the type and the value for the known relation
 * @returns An object containing RQL the query parts
 */
export const getRelationsSubjectObjectRqlParts = (
  entitySchema: EntitySchema,
  variableKeyAssociation: VariableKeyAssociation,
  sortRelation?: string,
  includeSubjectRelations = true,
  includeObjectRelations = true,
  knownRelation?: { relation: string; value: unknown },
  resolveRelations?: { [key: string]: true }
): RelationsRqlRequestParamsType => {
  const emptyObject = {
    selectionTransformers: [],
    selection: [],
    groupBy: [],
    restrictions: [],
    relationsTypes: [],
    sortVariable: null,
  };
  const subjectRelationsRqlRequestParams = includeSubjectRelations
    ? getRelationsRqlParts(
        entitySchema,
        true,
        variableKeyAssociation,
        sortRelation,
        knownRelation,
        resolveRelations
      )
    : emptyObject;
  const objectRelationsRqlRequestParams = includeObjectRelations
    ? getRelationsRqlParts(
        entitySchema,
        false,
        variableKeyAssociation,
        sortRelation,
        knownRelation,
        resolveRelations
      )
    : emptyObject;

  return {
    selectionTransformers: [
      ...subjectRelationsRqlRequestParams.selectionTransformers,
      ...objectRelationsRqlRequestParams.selectionTransformers,
    ],
    selection: [
      ...subjectRelationsRqlRequestParams.selection,
      ...objectRelationsRqlRequestParams.selection,
    ],
    groupBy: [
      ...subjectRelationsRqlRequestParams.groupBy,
      ...objectRelationsRqlRequestParams.groupBy,
    ],
    restrictions: [
      ...subjectRelationsRqlRequestParams.restrictions,
      ...objectRelationsRqlRequestParams.restrictions,
    ],
    relationsTypes: [
      ...subjectRelationsRqlRequestParams.relationsTypes,
      ...objectRelationsRqlRequestParams.relationsTypes,
    ],
    sortVariable:
      subjectRelationsRqlRequestParams.sortVariable != null
        ? subjectRelationsRqlRequestParams.sortVariable
        : objectRelationsRqlRequestParams.sortVariable,
  };
};

/**
 * Creates RQL query parts to fetch ids of entities linked by a relation to the given entity schema.
 *
 * @param entitySchema The entity schema abstraction class
 * @param isSubject True to fetch relations with this entity as subject, false for object
 * @param sortRelation The relation type used to sort
 * @param knownRelation An object containing the type and the value for the known relation
 * @returns An object containing RQL the query parts
 */
export const getRelationsRqlParts = (
  entitySchema: EntitySchema,
  isSubject: boolean,
  variableKeyAssociation: VariableKeyAssociation,
  sortRelation?: string,
  knownRelation?: { relation: string; value: unknown },
  resolveRelations?: { [key: string]: true }
): RelationsRqlRequestParamsType => {
  const selectionTransformers: Array<SelectionTransformer> = [];
  const selection: string[] = [];
  const groupBy: string[] = [];
  const restrictions: string[] = [];
  const relationsTypes: string[] = [];
  let sortVariable: string | null = null;

  const relationList = isSubject
    ? getCompatibleRelations(entitySchema.subjectRelationDefinitions)
    : getCompatibleRelations(entitySchema.objectRelationsDefinitions);

  // filter relations regarding fetchRelations definition
  const concernedRelations =
    resolveRelations === undefined
      ? relationList
      : relationList.filter((r) => r.type in Object.keys(resolveRelations));

  concernedRelations.forEach((relation) => {
    const relationType = getRelationKey(
      relation,
      isSubject ? "subject" : "object"
    );
    relationsTypes.push(relationType);

    const variable = getRelationVariable(
      relationType,
      "relation",
      variableKeyAssociation
    );
    const cardinality = relation.cardinality as CardinalityPair;
    // If this entity is the object of the relation,
    // we need to look at the second cardinality pair
    const cardinalityIndex = isSubject ? 0 : 1;

    if (
      cardinality[cardinalityIndex] === "*" ||
      cardinality[cardinalityIndex] === "+"
    ) {
      selection.push(`GROUP_CONCAT(${variable})`);
      selectionTransformers.push(parseEidList as SelectionTransformer);
    } else {
      selection.push(variable);
      selectionTransformers.push(identity as SelectionTransformer);
      groupBy.push(variable);
    }
    if (isSubject) {
      if (knownRelation != null && knownRelation.relation === relationType) {
        restrictions.push(
          `X ${relation.type} ${variable}?, ${variable} is ${relation.object.type}, M eid ${knownRelation.value}, X ${relation.type} M`
        );
      } else {
        restrictions.push(
          `X ${relation.type} ${variable}?, ${variable} is ${relation.object.type}`
        );
      }
    } else {
      if (knownRelation != null && knownRelation.relation === relationType) {
        restrictions.push(
          `${variable}? ${relation.type} X, ${variable} is ${relation.subject.type}, M eid ${knownRelation.value}, M ${relation.type} X`
        );
      } else {
        restrictions.push(
          `${variable}? ${relation.type} X, ${variable} is ${relation.subject.type}`
        );
      }
    }
    if (relationType === sortRelation) {
      sortVariable = variable;
    }
  });

  return {
    selectionTransformers,
    selection,
    groupBy,
    restrictions,
    relationsTypes,
    sortVariable,
  };
};

/**
 * Creates RQL query parts to fetch the value of attributes.
 *
 * @param attributesNames The names of the attributes to fetch
 * @param sortRelation The relation type used to sort
 * @param knownRelation An object containing the type and the value for the known relation
 * @returns An object containing RQL the query parts
 */
export const getAttributesRqlParts = (
  attributesNames: Array<string>,
  variableKeyAssociation: VariableKeyAssociation,
  sortRelation?: string,
  knownRelation?: { relation: string; value: unknown }
): AttributesRqlRequestParamsType => {
  const selectionTransformers: Array<SelectionTransformer> = [];
  const selection: string[] = [];
  const groupBy: string[] = [];
  const restrictions: string[] = [];
  let sortVariable: string | null = null;
  attributesNames.forEach((key) => {
    const iskeyEid = key === "X";
    let variable = key;
    if (!iskeyEid) {
      variable = getRelationVariable(key, "attribute", variableKeyAssociation);
    }
    selection.push(variable);
    selectionTransformers.push(identity as SelectionTransformer);
    groupBy.push(variable);
    if (!iskeyEid && knownRelation != null && knownRelation.relation === key) {
      restrictions.push(`X ${key} ${knownRelation.value}`);
    } else if (!iskeyEid) {
      restrictions.push(`X ${key} ${variable}`);
    }
    if (
      key === sortRelation ||
      (iskeyEid && (sortRelation === "id" || sortRelation === "eid"))
    ) {
      sortVariable = variable;
    }
  });

  return {
    selectionTransformers,
    selection,
    groupBy,
    restrictions,
    sortVariable,
  };
};

/**
 * Identity function.
 * Use this as a selection transformer if you do not want
 * to modify the data before showing it in the UI.
 *
 * @param x
 * @returns x
 */
export const identity: SelectionTransformer = <T>(x: T) => {
  return x;
};

/**
 * Converts an eid CSV list to a JavaScript int array
 *
 * @param eidsAsCSV Comma separated eid list
 * @returns
 */
export const parseEidList = (eidsAsCSV: string) =>
  eidsAsCSV === ""
    ? []
    : eidsAsCSV.split(", ").map((eidStr) => Number.parseInt(eidStr, 10));

/**
 * Generates the RQL selection part of the request, beggining with the Any keyword.
 *
 * @param selection An array of RQL variables to select
 * @throws Error when the given selection array is empty
 * @returns
 */
export const getSelectionClause = (selection: string[]) => {
  if (selection.length > 0) {
    return `Any ${selection.join(", ")}`;
  } else {
    throw new Error("RQL selection is empty");
  }
};

/**
 * Generates the RQL restriction part of the request, beggining with the WHERE keyword.
 *
 * @param restrictions An array of RQL restrictions
 * @param entityType The entity type to set for the request using 'X is entityType'. Use this to fetch lists of entities when the id is not specified.
 * @returns The restriction string, or an empty string if no restrictions were given
 */
export const getRestrictionsClause = (
  restrictions: string[],
  entityType?: string
) => {
  const restrictionsList = entityType
    ? [`X is ${entityType}`, ...restrictions]
    : restrictions;
  return restrictions.length > 0 ? "WHERE " + restrictionsList.join(", ") : "";
};

/**
 * Creates the RQL clause to group results using the given variables
 *
 * @param groupBy The list of variables used to group by
 * @returns
 */
export const getGroupByClause = (groupBy: string[]) => {
  return groupBy.length > 0 ? `GROUPBY ${groupBy.join(", ")}` : "";
};
